/*
  ======================================================================
   demo.c --- objekt in 2d
  ----------------------------------------------------------------------
   
	

  =======================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>

/* macros */

#define IX(i,j) ((i)+(N+2)*(j))

#define WINDKANAL 1
#define DEBUG 1
#define OBJEKT 1

/* external definitions (from solver.c) */

extern void dens_step ( int N, float * x, float * x0, float * u, float * v, float diff, float dt );
extern void vel_step ( int N, float * u, float * v, float * u0, float * v0, float visc, float dt );

/* global variables */

static int N;
static float dt, diff, visc;
static float force, source;
static int dvel;

static float * u, * v, * u_prev, * v_prev;
static float * dens, * dens_prev;
static int * object;

static int object_gegriffen = 0;
int obj_x, obj_y;
int obj_s_x, obj_s_y;

static int win_id;
static int win_x, win_y;
static int mouse_down[3];
static int omx, omy, mx, my, mpax, mpay;

#include "solver.c"


/*
  ----------------------------------------------------------------------
   free/clear/allocate simulation data
  ----------------------------------------------------------------------
*/


static void free_data ( void )
{
	if ( u ) free ( u );
	if ( v ) free ( v );
	if ( u_prev ) free ( u_prev );
	if ( v_prev ) free ( v_prev );
	if ( dens ) free ( dens );
	if ( dens_prev ) free ( dens_prev );
	if ( object ) free ( object );
}

void make_quader()
{
	int x,y;
	//int xlow=20, xhig=25;
	//int ylow=50, yhig=59;
	int xlow = obj_x - obj_s_x / 2;
	int xhig = obj_x + obj_s_x / 2;
	int ylow = obj_y - obj_s_y / 2;
	int yhig = obj_y + obj_s_y / 2;

	int i, size = (N+2)*(N+2);
	for ( i=0 ; i<size ; i++ ) {
		object[i] = 0;
	}

	for ( x = xlow ; x<xhig ; x ++) {
		for ( y=ylow ; y<yhig ; y++) {
			object[IX(x,y)] = 2;			// border of object = 2
		}
	}	
	for ( x = xlow+1 ; x<xhig-1 ; x ++) {
		for ( y=ylow+1 ; y<yhig-1 ; y++) {
			object[IX(x,y)] = 1;			// inner part of object = 1
		}
	}
}

float dist2(float x1, float y1, float x2, float y2 )
{
	return (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1);
}

void make_circle()
{
	int x,y;
	float radius = 5.7;
	float radius2 = radius*radius;	
	for ( x=10 ; x<=N-10 ; x++ ) {
		for ( y=10 ; y<=N-10 ; y++ ) {
			if ( dist2 (x,y, obj_x, obj_y ) < radius2 ) object[IX(x,y)] = 2;
		}
	}

	radius = 4;
	radius2 = radius*radius;	
	for ( x=1 ; x<=N ; x++ ) {
		for ( y=1 ; y<=N ; y++ ) {
			if ( dist2 (x,y, obj_x, obj_y ) < radius2 ) object[IX(x,y)] = 1;
		}
	}


}

void make_object_wk()
{
	int x,y;
	int xlow=20, xhig=25;
	int ylow=50, yhig=59;
	for ( x = xlow ; x<xhig ; x ++) {
		for ( y=ylow ; y<yhig ; y++) {
			object[IX(x,y)] = 2;			// border of object = 2
		}
	}	
	for ( x = xlow+1 ; x<xhig-1 ; x ++) {
		for ( y=ylow+1 ; y<yhig-1 ; y++) {
			object[IX(x,y)] = 1;			// inner part of object = 1
		}
	}	
}

void make_object()
{
#if OBJEKT == 1
	make_quader();
#endif
#if OBJEKT == 2
	make_circle();
#endif
#if OBJEKT == 3
	make_object_wk();
#endif
}



static void clear_data ( void )
{
	int i, size=(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = u_prev[i] = v_prev[i] = dens[i] = dens_prev[i] = 0.0f;
		object[i] = 0;
	}
	make_object();
}

static int allocate_data ( void )
{
	int size = (N+2)*(N+2);

	u			= (float *) malloc ( size*sizeof(float) );
	v			= (float *) malloc ( size*sizeof(float) );
	u_prev		= (float *) malloc ( size*sizeof(float) );
	v_prev		= (float *) malloc ( size*sizeof(float) );
	dens		= (float *) malloc ( size*sizeof(float) );	
	dens_prev	= (float *) malloc ( size*sizeof(float) );
	object		= (int *)   malloc ( size*sizeof(int  ) );

	if ( !u || !v || !u_prev || !v_prev || !dens || !dens_prev || !object) {
		fprintf ( stderr, "cannot allocate data\n" );
		return ( 0 );
	}

	return ( 1 );
}


/*
  ----------------------------------------------------------------------
   OpenGL specific drawing routines
  ----------------------------------------------------------------------
*/

static void pre_display ( void )
{
	glViewport ( 0, 0, win_x, win_y );
	glMatrixMode ( GL_PROJECTION );
	glLoadIdentity ();
	gluOrtho2D ( 0.0, 1.0, 0.0, 1.0 );
	glClearColor ( 0.0f, 0.0f, 0.0f, 1.0f );
	glClear ( GL_COLOR_BUFFER_BIT );
}

static void post_display ( void )
{
	glutSwapBuffers ();
}

static void draw_velocity ( void )
{
	int i, j;
	float x, y, h;

	h = 1.0f/N;

	glColor4f ( 0.7f, 0.0f, 0.0f, 0.2f );
	glLineWidth ( 1.0f );

	glBegin ( GL_LINES );

		for ( i=1 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=1 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;

				glVertex2f ( x, y );
				glVertex2f ( x+u[IX(i,j)], y+v[IX(i,j)] );
			}
		}

	glEnd ();
}

static void draw_density ( void )
{
	int i, j;
	float x, y, h, d00, d01, d10, d11;

	h = 1.0f/N;

	glBegin ( GL_QUADS );

		for ( i=0 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=0 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;

				d00 = dens[IX(i,j)];
				d01 = dens[IX(i,j+1)];
				d10 = dens[IX(i+1,j)];
				d11 = dens[IX(i+1,j+1)];

				glColor3f ( d00, d00, d00 ); glVertex2f ( x, y );
				glColor3f ( d10, d10, d10 ); glVertex2f ( x+h, y );
				glColor3f ( d11, d11, d11 ); glVertex2f ( x+h, y+h );
				glColor3f ( d01, d01, d01 ); glVertex2f ( x, y+h );

				if ( object[IX(i,j)] >0 ) {
					if ( object[IX(i,j)] == 1 ) glColor4f ( 1, 1, 0, 0.3 );
					if ( object[IX(i,j)] == 2 ) glColor4f ( 0, 1, 1, 0.3 );
					glVertex2f ( x, y );
					glVertex2f ( x+h, y );
					glVertex2f ( x+h, y+h );
					glVertex2f ( x, y+h );
				}

			}
		}

	glEnd ();
}

/*
  ----------------------------------------------------------------------
   relates mouse movements to forces sources
  ----------------------------------------------------------------------
*/

static void get_from_UI_alt ( float * d, float * u, float * v )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}

	if ( !mouse_down[0] && !mouse_down[2] ) return;

	i = (int)((       mx /(float)win_x)*N+1);
	j = (int)(((win_y-my)/(float)win_y)*N+1);

	if ( i<1 || i>N || j<1 || j>N ) return;

	if ( mouse_down[0] ) {
		u[IX(i,j)] = force * (mx-omx);
		v[IX(i,j)] = force * (omy-my);
	}

	if ( mouse_down[2] ) {
		d[IX(i,j)] = source;
	}

	omx = mx;
	omy = my;

	return;
}



static void get_from_UI ( float * d, float * u, float * v )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}

	i = (int)((       mx /(float)win_x)*N+1);
	j = (int)(((win_y-my)/(float)win_y)*N+1);

	if ( object_gegriffen == 1 )
	{
		obj_x = obj_x + (i-mpax);
		obj_y = obj_y + (j-mpay);
		make_object();
	}

	mpax = i;
	mpay = j;

	if ( !mouse_down[0] ) object_gegriffen = 0;

	if ( !mouse_down[0] && !mouse_down[2] ) return;

	if ( i<1 || i>N || j<1 || j>N ) return;

	if ( mouse_down[0] ) {
		if ( object[IX(i,j)] > 0 ) {
			object_gegriffen = 1;
		}
		else
		{
			u[IX(i,j)] = force * (mx-omx);
			v[IX(i,j)] = force * (omy-my);
		}
	}

	if ( mouse_down[2] ) {
		d[IX(i,j)] = source;
	}

	omx = mx;
	omy = my;

	return;
}


static void ventilator ( float * d, float * u, float * v )
{
	int i, j;


	for ( i = 1; i <= N; i++ ) {
	  for ( j=1; j<3; j++ ) 
		u[IX(j,i)] += 1.1f;
	}


	for ( i=N/2-4; i<N/2+4; i+=3) 
	   d[IX(1,i)] = source/5;

	omx = mx;
	omy = my;

	return;
}



/*
  ----------------------------------------------------------------------
   GLUT callback routines
  ----------------------------------------------------------------------
*/

static void key_func ( unsigned char key, int x, int y )
{
	switch ( key )
	{
		case 'c':
		case 'C':
			clear_data ();
			break;

		case 'q':
		case 'Q':
			free_data ();
			exit ( 0 );
			break;

		case 'v':
		case 'V':
			dvel = !dvel;
			break;
	}
}

static void mouse_func ( int button, int state, int x, int y )
{
	omx = mx = x;
	omx = my = y;

	mouse_down[button] = state == GLUT_DOWN;
}

static void motion_func ( int x, int y )
{
	mx = x;
	my = y;
}

static void reshape_func ( int width, int height )
{
	glutSetWindow ( win_id );
	glutReshapeWindow ( width, height );

	win_x = width;
	win_y = height;
}

static void idle_func ( void )
{
	get_from_UI ( dens_prev, u_prev, v_prev );
#ifdef WINDKANAL	
	ventilator ( dens_prev, u_prev, v_prev );
#endif
	vel_step ( N, u, v, u_prev, v_prev, visc, dt );
	dens_step ( N, dens, dens_prev, u, v, diff, dt );

	glutSetWindow ( win_id );
	glutPostRedisplay ();
}

static void display_func ( void )
{
	pre_display ();
		
		draw_density ();
		if ( dvel ) draw_velocity ();

	post_display ();
}


/*
  ----------------------------------------------------------------------
   open_glut_window --- open a glut compatible window and set callbacks
  ----------------------------------------------------------------------
*/

static void open_glut_window ( void )
{
	glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE );

	glutInitWindowPosition ( 0, 0 );
	glutInitWindowSize ( win_x, win_y );
	win_id = glutCreateWindow ( "Alias | wavefront" );

	glClearColor ( 0.0f, 0.0f, 0.0f, 1.0f );
	glClear ( GL_COLOR_BUFFER_BIT );
	glutSwapBuffers ();
	glClear ( GL_COLOR_BUFFER_BIT );
	glutSwapBuffers ();

	glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	glEnable(GL_BLEND);

	pre_display ();

	glutKeyboardFunc ( key_func );
	glutMouseFunc ( mouse_func );
	glutMotionFunc ( motion_func );
	glutReshapeFunc ( reshape_func );
	glutIdleFunc ( idle_func );
	glutDisplayFunc ( display_func );
}


/*
  ----------------------------------------------------------------------
   main --- main routine
  ----------------------------------------------------------------------
*/

int main ( int argc, char ** argv )
{
	glutInit ( &argc, argv );

	if ( argc != 1 && argc != 6 ) {
		fprintf ( stderr, "usage : %s N dt diff visc force source\n", argv[0] );
		fprintf ( stderr, "where:\n" );\
		fprintf ( stderr, "\t N      : grid resolution\n" );
		fprintf ( stderr, "\t dt     : time step\n" );
		fprintf ( stderr, "\t diff   : diffusion rate of the density\n" );
		fprintf ( stderr, "\t visc   : viscosity of the fluid\n" );
		fprintf ( stderr, "\t force  : scales the mouse movement that generate a force\n" );
		fprintf ( stderr, "\t source : amount of density that will be deposited\n" );
		exit ( 1 );
	}

	if ( argc == 1 ) {
		N = 128;
		dt = 0.1f;
		diff = 0.0f;
		visc = 0.0f;
		force = 5.0f;
		source = 100.0f;
		fprintf ( stderr, "Using defaults : N=%d dt=%g diff=%g visc=%g force = %g source=%g\n",
			N, dt, diff, visc, force, source );
	} else {
		N = atoi(argv[1]);
		dt = atof(argv[2]);
		diff = atof(argv[3]);
		visc = atof(argv[4]);
		force = atof(argv[5]);
		source = atof(argv[6]);
	}

	printf ( "\n\nHow to use this demo:\n\n" );
	printf ( "\t Add densities with the right mouse button\n" );
	printf ( "\t Add velocities with the left mouse button and dragging the mouse\n" );
	printf ( "\t Toggle density/velocity display with the 'v' key\n" );
	printf ( "\t Clear the simulation by pressing the 'c' key\n" );
	printf ( "\t Quit by pressing the 'q' key\n" );

	dvel = 0;

	if ( !allocate_data () ) exit ( 1 );

#ifdef OBJEKT
	obj_x = N / 8;
	obj_y = N / 2;
	obj_s_x = N / 20;
	obj_s_y = N / 12;	
#endif

	clear_data ();


	win_x = 512;
	win_y = 512;
	open_glut_window ();

	glutMainLoop ();

	exit ( 0 );
}