
#include <windows.h>
#include "tgaload.h"
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h> 	/* for test_bit */

#define BIN			2

/* Pointer to OpenGL extension function  */

PFNGLCOLORTABLEEXTPROC glColorTableEXT;

/*	Check whether a bit in an integer is 0 or 1
 */
tgaFLAG test_bit ( int Bit, tgaFLAG FlagVar )
{
	if ( FlagVar & tgaFLAG ( pow ( BIN,Bit ) ) )
		return 1;
	else
		return 0;
}


/* Fetch the address of the function 'glColorTableEXT' from the active opengl
 * renderer.  This will probably cause an access violation if the driver
 * doesn't support colour palettes.  Maybe I should check first  :)
 */
void tgaGetColorEXT ( void )
{
   glColorTableEXT = ( PFNGLCOLORTABLEEXTPROC ) wglGetProcAddress ( "glColorTableEXT" );
}



void tgaChangePalette ( unsigned char *palette )

{
	glColorTableEXT ( GL_TEXTURE_2D, GL_RGB, 256, GL_RGB, GL_UNSIGNED_BYTE, palette );
}



void tgaSetTexParams  ( unsigned int filter_type, unsigned int application )

{

   glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter_type );

   glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter_type );

   /*  Set Texture Evironment */
   glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, (float) application );

}



unsigned char *tgaGetPalette ( tgaHeader_t info, FILE *file )

{

	unsigned char  *palette;

   unsigned char  temp;

   int    			components;

   int				bytes;



   // Check we have a 256 colour palette of 8-bit values (std. output from Photoshop)

	if ( info.colour_map_length != 256    || 
		info.colour_map_entry_size != 24 ||
        info.pixel_depth != 8 )

   	return 0;


   components = info.colour_map_entry_size / info.pixel_depth;

   bytes      = info.colour_map_length * components;


   palette = (unsigned char*) malloc ( bytes );


   if ( palette == NULL )

   	return 0;


   fread ( palette, sizeof (unsigned char), bytes, file);


   /* Palette is stored as BGR, fiddle it to RGB */

   for ( int i = 0; i < bytes; i += components )

   {
   	temp = palette[i];
      palette[i] = palette[i + 2];
      palette[i + 2] = temp;
   }


   return palette;

}



unsigned char *tgaGetImageData ( tgaHeader_t info, FILE *file )

{

	unsigned char  *data;

	unsigned char  temp;

   int    			components;

   int				bytes;


   components = info.pixel_depth / 8;

   bytes      = info.width * info.height * components;


   data = (unsigned char*) malloc ( bytes );


   if ( data == NULL )

   	return 0;


   fread ( data, sizeof (unsigned char), bytes, file);


   /* Image is stored as BGR(A), make it RGB(A)     */

   for (int i = 0; i < bytes; i += components )

   {
       temp = data[i];
       data[i] = data[i + 2];
       data[i + 2] = temp;
   }

   return data;

}



void tgaPassImage ( image_t *p, GLenum colour_type )

{
   /*  Set OpenGL Texture Image */
   if ( colour_type == GL_COLOR_INDEX )
	{
	 	glColorTableEXT ( GL_TEXTURE_2D, GL_RGB, p->info.colour_map_length, GL_RGB,GL_UNSIGNED_BYTE, p->palette );
		gluBuild2DMipmaps ( GL_TEXTURE_2D, GL_COLOR_INDEX8_EXT, p->info.width,
							p->info.height, colour_type, GL_UNSIGNED_BYTE,
                   		    p->data );
   }
   else
		gluBuild2DMipmaps ( GL_TEXTURE_2D, colour_type, p->info.width,
		 				  p->info.height, colour_type, GL_UNSIGNED_BYTE,
                   		  p->data );
}



void tgaInit ( image_t *p )
{
	p->palette = NULL;
   p->data    = NULL;
}


void tgaFree ( image_t *p )
{
	if ( p->data != NULL )
   	free ( p->data );
   if ( p->palette != NULL )
   	free ( p->palette );
}


void tgaError ( char *error_string, char *file_name, image_t *p )
{
	printf  ( "%s - %s type:%i depth:%i \n", error_string, file_name,p->info.image_type,p->info.pixel_depth );
   tgaFree ( p );
}

void tgaInfo ( image_t *p )
{
	printf  ( "id_length:%i  \n",p->info.id_length);
	printf  ( "colour_map_type:%i  \n",p->info.colour_map_type);
	printf  ( "type:%i  \n",p->info.image_type);
	
	printf  ( "colour_map_first_entry:%i  \n",p->info.colour_map_first_entry);
	printf  ( "colour_map_length:%i  \n",p->info.colour_map_length);
	printf  ( "colour_map_entry_size:%i  \n",p->info.colour_map_entry_size);

	printf  ( "x_origin:%i \n",p->info.x_origin );
	printf  ( "y_origin:%i \n",p->info.y_origin );
	printf  ( "width:%i \n",p->info.width );
	printf  ( "height:%i \n",p->info.height );
	printf  ( "depth:%i \n",p->info.pixel_depth );

}


int tgaLoadImage ( char *file_name, image_t *p, tgaFLAG mode )
{
	FILE 			  *file;

 	if (( file = fopen ( file_name, "rb" )) == NULL )
   {
  		printf ( "File Not Found : %s\n", file_name );
//      fclose ( file );
      return 0;
   }

   tgaInit ( p );

   fread ( &p->info, ( sizeof ( tgaHeader_t )), 1, file );
	tgaInfo(p);
   /* Recognise colour mapped (paletted) image) */
   if ( p->info.image_type == 1 )
	{
   	p->palette = tgaGetPalette   ( p->info, file );
      p->data    = tgaGetImageData ( p->info, file );
    	tgaPassImage ( p, GL_COLOR_INDEX );

   } else

   /* Recognise RGB(A) true colour image */
   if ( p->info.image_type == 2 )
   {
		p->data = tgaGetImageData ( p->info, file );
      if ( p->info.pixel_depth == 24 )
			tgaPassImage ( p, GL_RGB );
      else if ( p->info.pixel_depth == 32 )
			tgaPassImage ( p, GL_RGBA );
      else
      	tgaError ( "Image is too deep", file_name, p );
   } else

   /* Recognise greyscale image */
   if ( p->info.image_type == 3 )
   {
		p->data = tgaGetImageData ( p->info, file );

      /* 8-bit greyscale only :) */
      if ( p->info.pixel_depth == 8 )
		{
      	/* Check if user has specified a 'forced mode' */
	   	if ( test_bit ( 3, mode ))
   			tgaPassImage ( p, GL_LUMINANCE );

      	else

     	 	if ( test_bit ( 2, mode ))

   			tgaPassImage ( p, GL_ALPHA );
      	else
			{
         	tgaError ( "Unspecified greyscale mode ", file_name, p );
            return 0;
         }
      }
   }

   /*  Image type not Recognised ! */
   else
   {
   	tgaError ( "Unable to determine image type", file_name, p );
	   return 0;   // Unsupported!
   }

   /*  Trap for mode == TGA_FREE  */
   if ( test_bit ( 0, mode ))
	   tgaFree ( p );

   fclose  ( file );
	return 1;
}

int tgaLoadPalette ( char *file_name, unsigned char *palette )
{
	FILE 			  *file;
   tgaHeader_t   info;

 	if (( file = fopen ( file_name, "rb" )) == NULL )
   {
  		printf ( "File Not Found : %s\n", file_name );
      fclose ( file );
      return 0;
   }

   fread ( &info, ( sizeof ( tgaHeader_t )), 1, file );

   if (info.image_type == 1 )    // If it has a palette, read it
   	palette = tgaGetPalette   ( info, file );
   else
   	return 0;   // Unsupported!

   fclose  ( file );
	return 1;
}
